# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Command-line interface for `invenio_moodle`."""

from click import STRING, group, option, secho
from click_params import URL
from flask.cli import with_appcontext
from invenio_config_tugraz import get_identity_from_user_by_email
from invenio_records_lom.proxies import current_records_lom
from marshmallow import ValidationError

from .api import fetch_moodle
from .types import Color


@group()
def moodle() -> None:
    """invenio-moodle commands."""


@moodle.command()
@option("--endpoint", type=URL, required=True)
@option("--user-email", type=STRING, required=True)
@with_appcontext
def import_by_endpoint(endpoint: str, user_email: str) -> None:
    """Fetch data from MOODLE_FETCH_URL and insert it into the database."""
    identity = get_identity_from_user_by_email(email=user_email)
    records_service = current_records_lom.records_service

    try:
        fetch_moodle(endpoint, records_service, identity)
    except ValidationError as error:
        secho(error, fg=Color.error)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Celery tasks for `invenio-moodle`."""

import traceback

from celery import shared_task
from flask import current_app
from flask_mail import Message

from .api import fetch_moodle


@shared_task(ignore_result=True)
def try_fetch_moodle_except_mail() -> None:
    """Fetch data from moodle and enter it into database."""
    try:
        moodle_fetch_url = current_app.config["MOODLE_FETCH_URL"]
        fetch_moodle(moodle_fetch_url)

    except Exception:  # noqa: BLE001
        config = current_app.config
        msg = Message(
            "Something went wrong when fetching data from moodle",
            sender=config["MOODLE_ERROR_MAIL_SENDER"],
            recipients=config["MOODLE_ERROR_MAIL_RECIPIENTS"],
            body=traceback.format_exc(),
        )
        current_app.extensions["mail"].send(msg)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Types."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto, unique
from functools import singledispatchmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from invenio_records_lom import LOMMetadata


@dataclass(frozen=True)
class Color:
    """The class is for the output color management."""

    neutral = "white"
    error = "red"
    warning = "yellow"
    abort = "magenta"
    success = "green"
    alternate = ("blue", "cyan")


@dataclass(frozen=True)
class FileCacheInfo:
    """Holds a file-path and the file's md5-hash."""

    hash_md5: str
    path: Path


@dataclass(frozen=True)
class Key(ABC):
    """Common ancestor to all Key classes."""

    @property
    @abstractmethod
    def resource_type(self) -> str:
        """The resource_type associated to the key, one of `LOM_RESOURCE_TYPES`."""

    @abstractmethod
    def __str__(self) -> str:
        """Convert `self` to unique string representation."""

    def __hash__(self) -> str:
        """Get hash."""
        return self.get_moodle_pid_value()

    @abstractmethod
    def get_moodle_pid_value(self) -> str:
        """Return the primary hash of Key."""


@dataclass(frozen=True)
class FileKey(Key):
    """Key for files as to disambiguate it from keys for units and courses."""

    url: str
    year: str
    semester: str
    hash_md5: str

    resource_type = "file"

    @singledispatchmethod
    def __init__(self, url: str, year: str, semester: str, hash_md5: str) -> None:
        """Construct."""
        # dataclass frozen is nice but needs following not handy construct
        object.__setattr__(self, "url", url)
        object.__setattr__(self, "year", year)
        object.__setattr__(self, "semester", semester)
        object.__setattr__(self, "hash_md5", hash_md5)

    @__init__.register
    def _(
        self,
        moodle_file_metadata: dict,
    ) -> None:
        """Create `cls` via info from moodle-json and file-cache."""
        url = moodle_file_metadata["fileurl"]
        year = moodle_file_metadata["year"]
        semester = moodle_file_metadata["semester"]
        hash_md5 = moodle_file_metadata["contenthash"]
        self.__init__(url, year, semester, hash_md5)

    def __str__(self) -> str:
        """Get string-representation."""
        url = f"url={self.url}"
        year = f"year={self.year}"
        semester = f"semester={self.semester}"
        hash_md5 = f"hash_md5={self.hash_md5}"
        return f"FileKey({url}, {year}, {semester}, {hash_md5})"

    def get_moodle_pid_value(self) -> str:
        """Get moodle pid value."""
        return self.hash_md5


@dataclass(frozen=True)
class LinkKey(Key):
    """Key for links only records."""

    url: str

    resource_type = "link"

    @singledispatchmethod
    def __init__(self, url: str) -> None:
        """Construct."""
        object.__setattr__(self, "url", url)

    def __str__(self) -> str:
        """Get string-representation."""
        url = f"url={self.url}"
        return f"LinkKey({url})"

    def get_moodle_pid_value(self) -> str:
        """Get moodle pid value."""
        return self.url


@unique
class Status(Enum):
    """Status."""

    NEW = auto()
    EDIT = auto()


@dataclass
class BaseRecord:
    """Base."""

    key: Key
    pid: str
    metadata: LOMMetadata
    status: Status

    @property
    def json(self) -> dict:
        """Get json."""
        return self.metadata.json


@dataclass
class FileRecord(BaseRecord):
    """File."""

    @property
    def url(self) -> str:
        """Get url."""
        return self.key.url

    @property
    def hash_md5(self) -> str:
        """Get hash_md5."""
        return self.key.hash_md5


@dataclass
class LinkRecord(BaseRecord):
    """Link."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

from collections.abc import Callable
from json import load
from pathlib import Path

import pytest
from flask import Flask

from invenio_moodle import InvenioMoodle


@pytest.fixture(scope="module")
def create_app(instance_path: str) -> Callable:
    """Application factory fixture."""

    def factory(**config: dict) -> Flask:
        app = Flask("testapp", instance_path=instance_path)
        app.config.update(**config)
        InvenioMoodle(app)
        return app

    return factory


@pytest.fixture(scope="module")
def minimal_record() -> dict:
    """Create minimal record."""
    filename = Path("data/minimal_record.json")
    directory = Path(__file__).parent
    filepath = directory / filename
    with filepath.open("r") as fp:
        return load(fp)


@pytest.fixture()
def expected_lom_metadata() -> None:
    """Expectd unit."""
    filename = Path("data/lom_metadata.json")
    directory = Path(__file__).parent
    filepath = directory / filename
    with filepath.open("r") as fp:
        return load(fp)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration for invenio-moodle."""

from flask_babelex import gettext as _
from invenio_rdm_records.services.pids.providers import ExternalPIDProvider

MOODLE_CELERY_BEAT_SCHEDULE = {}
"""The celery beat schedule.

This variable is used to configure the celery task which is used to import data
from moodle on a regularly basis.
"""

MOODLE_PERSISTENT_IDENTIFIER_PROVIDERS = [
    ExternalPIDProvider("moodle", "moodle", label=_("MOODLE ID")),
]
"""List of persistent identifier providers.

The values are added to the LOM_PERSISTENT_IDENTIFIER_PROVIDERS list.
"""

MOODLE_PERSISTENT_IDENTIFIERS = {
    "moodle": {
        "providers": ["moodle"],
        "required": False,
        "label": _("MOODLE"),
    },
}
"""Dict of persistent identifiers.

The values are added to the LOM_PERSISTENT_IDENTIFIERS dict.
"""

MOODLE_FETCH_URL = ""
"""The url of the moodle endpoint from where should be fetched the metadata."""

MOODLE_ERROR_MAIL_SENDER = ""
"""The variable is used to configure the sender of the error emails."""

MOODLE_ERROR_MAIL_RECIPIENTS = []
"""The variable is used to configure the list of recipients of the error emails."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Provides API for Moodle."""

from .ext import InvenioMoodle

__version__ = "0.2.2"

__all__ = ("__version__", "InvenioMoodle")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Utilities for inserting moodle-data into invenio-style database."""

from __future__ import annotations

from typing import TYPE_CHECKING

from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_pidstore.models import PersistentIdentifier
from invenio_records_lom.utils import LOMMetadata

from .convert import convert_moodle_to_lom
from .files import add_file_to_draft
from .types import BaseRecord, FileKey, FileRecord, LinkKey, LinkRecord, Status

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_lom import LOMRecordService
    from invenio_records_resources.services.records import RecordItem

    from .types import Key


def is_not_moodle_only_course(moodle_course_metadata: dict) -> bool:
    """Check if it is a moodle only course."""
    return moodle_course_metadata["courseid"] != "0"


def is_course_root(sourceid: str) -> bool:
    """Check if parent exists."""
    return sourceid == "-1"


def extract_moodle_records(moodle_data: dict) -> list[dict]:
    """Create moodle file jsons."""
    return [
        file_json
        for _, moodle_course in moodle_data["moodlecourses"].items()
        for file_json in moodle_course["files"]
    ]


def create_draft(
    key: Key,
    moodle_pid_value: str,
    records_service: LOMRecordService,
    identity: Identity,
) -> RecordItem:
    """Create draft with empty metadata."""
    pids_dict = {
        "moodle": {
            "provider": "moodle",
            "identifier": moodle_pid_value,
        },
    }
    metadata = LOMMetadata.create(resource_type=key.resource_type, pids=pids_dict)
    metadata.append_identifier(moodle_pid_value, catalog="moodle")
    return records_service.create(data=metadata.json, identity=identity)


def get_from_database_or_create(
    key: Key,
    records_service: LOMRecordService,
    identity: Identity,
) -> BaseRecord:
    """Fetch moodle-result corresponding to `key`, create database-entry if none exists.

    :param Key key: the key which to attempt fetching from pidstore


    File:
    A File could be already in database. A File should not be added twice and it
    is looked up by its file hash (hash_md5). If it is there already it will be
    linked to the new Unit where it is used.

    Unit:
    An Unit will ever be created from scratch. They are unique and created every
    year new

    Course:
    A Course could be new, but it is usually not. A Course only changes if a new
    curriculum will be defined.
    """
    moodle_pid_value = key.get_moodle_pid_value()

    try:
        moodle_pid = PersistentIdentifier.get(
            pid_type="moodle",
            pid_value=moodle_pid_value,
        )
    except PIDDoesNotExistError:
        draft = create_draft(key, moodle_pid_value, records_service, identity)
        pid: str = draft.id
        lom_metadata = LOMMetadata(draft.to_dict(), overwritable=True)
        status = Status.NEW
    else:
        # get lomid corresponding to moodle_pid
        lom_pid = PersistentIdentifier.get_by_object(
            pid_type="lomid",
            object_type=moodle_pid.object_type,
            object_uuid=moodle_pid.object_uuid,
        )

        pid: str = lom_pid.pid_value
        metadata = records_service.edit(id_=pid, identity=identity).to_dict()
        lom_metadata = LOMMetadata(metadata, overwritable=True)
        status = Status.EDIT

    if isinstance(key, FileKey):
        type_of_record = FileRecord
    elif isinstance(key, LinkKey):
        type_of_record = LinkRecord
    else:
        type_of_record = BaseRecord

    return type_of_record(key, pid, lom_metadata, status)


def build_intermediate_records(
    moodle_records: list[dict],
    records_service: LOMRecordService,
    identity: Identity,
) -> dict[Key, BaseRecord]:
    """Build course tree.

    :param dict moodle_data: Data whose format matches `MoodleSchema`
    """
    records = {}

    # prepare: gather necessary information, create records if no
    # previous versions exist
    for moodle_file_metadata in moodle_records:
        file_key = FileKey(moodle_file_metadata)
        file_record = get_from_database_or_create(file_key, records_service, identity)
        convert_moodle_to_lom(moodle_file_metadata, file_record.metadata)
        records[file_key] = file_record

    return records


def import_record(
    record: BaseRecord,
    records_service: LOMRecordService,
    identity: Identity,
) -> None:
    """Import Record."""
    if isinstance(record, FileRecord) and record.status == Status.NEW:
        add_file_to_draft(record, records_service.draft_files, identity)

    records_service.update_draft(
        id_=record.pid,
        data=record.json,
        identity=identity,
    )
    records_service.publish(id_=record.pid, identity=identity)


def remove_moodle_only_course(moodle_records: dict) -> None:
    """Remove moodle only course.

    This is indicated by the courseid == 0
    """
    for moodle_file_metadata in moodle_records:
        not_moodle_only_courses = list(
            filter(is_not_moodle_only_course, moodle_file_metadata["courses"]),
        )
        moodle_file_metadata["courses"] = not_moodle_only_courses


def post_processing(moodle_records: dict) -> dict:
    """Post process moodle data.

    Remove unwanted fields:
        - remove course with courseid == 0
    """
    remove_moodle_only_course(moodle_records)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Flask extension for invenio-moodle."""

from flask import Flask

from . import config


class InvenioMoodle:
    """invenio-moodle extension."""

    def __init__(self, app: Flask = None) -> None:
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        self.init_config(app)
        app.extensions["invenio-moodle"] = self

    def init_config(self, app: Flask) -> None:
        """Initialize configuration."""
        app.config.setdefault("LOM_PERSISTENT_IDENTIFIERS", {})
        app.config.setdefault("LOM_PERSISTENT_IDENTIFIER_PROVIDERS", [])

        for key in dir(config):
            value = getattr(config, key)
            lom_config_variable = key.replace("MOODLE_", "LOM_")

            if lom_config_variable in app.config:
                container = app.config[lom_config_variable]

                if isinstance(container, list):
                    container.extend(value)
                elif isinstance(container, dict):
                    container.update(value)
                else:
                    container = value

            elif key.startswith("MOODLE_"):
                app.config.setdefault(key, value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Schemas for validating input from moodle."""

from collections import Counter, defaultdict

from invenio_records_lom.services.schemas.fields import ControlledVocabularyField
from marshmallow import Schema, ValidationError, validates_schema
from marshmallow.fields import Constant, Dict, List, Nested, String


class ClassificationValuesSchema(Schema):
    """Moodle classification-values schema."""

    identifier = String(required=True)
    name = String(required=True)


class ClassificationSchema(Schema):
    """Moodle classification schema."""

    type = Constant("oefos", required=True)  # noqa: A003
    url = Constant(
        "https://www.data.gv.at/katalog/dataset/stat_ofos-2012",
        required=True,
    )
    values = List(Nested(ClassificationValuesSchema), required=True)


class CourseSchema(Schema):
    """Moodle course schema."""

    courseid = String(required=True)
    courselanguage = String(required=True)
    coursename = String(required=True)
    description = String(required=True)
    identifier = String(required=True)
    lecturer = String(required=True)
    objective = String(required=True)
    organisation = String(required=True)
    sourceid = String(required=True)
    structure = ControlledVocabularyField(
        vocabulary=[
            "",
            "Seminar (SE)",
            "Vorlesung (VO)",
            "Übung (UE)",
            "Seminar-Projekt (SP)",
            "Vorlesung und Übung (VU)",
            "Orientierungslehrveranstaltung (OL)",
        ],
        required=True,
    )


class LicenseSchema(Schema):
    """Moodle license schema."""

    fullname = String(required=True)
    shortname = String(required=True)
    source = String(required=True)


class PersonSchema(Schema):
    """Moodle person schema."""

    firstname = String(required=True)
    lastname = String(required=True)
    role = String(required=True)


class FileSchema(Schema):
    """Moodle file schema."""

    abstract = String(required=True)
    classification = List(Nested(ClassificationSchema), required=True)
    contenthash = String(required=True)
    context = String(required=True)
    courses = List(Nested(CourseSchema), required=True)
    filecreationtime = String(required=True)
    filesize = String(required=True)
    fileurl = String(required=True)
    language = String(required=True)
    license = Nested(LicenseSchema)  # noqa: A003
    mimetype = String(required=True)
    persons = List(Nested(PersonSchema), required=True)
    resourcetype = String(required=True)
    semester = ControlledVocabularyField(vocabulary=["SS", "WS"], required=True)
    tags = List(String(), required=True)
    timereleased = String(required=True)
    title = String(required=True)
    year = String(required=True)


class MoodleCourseSchema(Schema):
    """Moodle moodlecourse schema."""

    files = List(Nested(FileSchema))


class MoodleSchema(Schema):
    """Moodle moodlecourses schema.

    Data coming from moodle should be in this format.
    """

    applicationprofile = String(required=True)
    moodlecourses = Dict(
        keys=String(),
        values=Nested(MoodleCourseSchema, required=True),
    )

    @validates_schema
    def validate_urls_unique(self, data: dict, **__: dict) -> None:
        """Check that each file-URL only appears once."""
        urls_counter = Counter(
            file_["fileurl"]
            for _, moodlecourse in data["moodlecourses"].items()
            for file_ in moodlecourse["files"]
        )
        duplicated_urls = [url for url, count in urls_counter.items() if count > 1]
        if duplicated_urls:
            msg = f"Different file-JSONs with same URL {duplicated_urls}."
            raise ValidationError(msg)

    @validates_schema
    def validate_course_jsons_unique_per_courseid(self, data: dict, **__: dict) -> None:
        """Validate against unique courseid.

        Check that course-ids that appear multiple times have same
        json in all their appearances.
        """
        jsons_by_courseid = defaultdict(list)
        for moodlecourse in data["moodlecourses"].values():
            for file_ in moodlecourse["files"]:
                for course in file_["courses"]:
                    course_id = course["courseid"]
                    if course not in jsons_by_courseid[course_id]:
                        jsons_by_courseid[course_id].append(course)

        ambiguous_courseids = {
            course_id
            for course_id, jsons in jsons_by_courseid.items()
            if len(jsons) > 1
        }

        # '0' is special courseid shared by all moodle-only courses,
        # it is allowed to be ambiguous
        ambiguous_courseids -= {"0"}

        if ambiguous_courseids:
            course_ids = ", ".join(course_id for course_id in ambiguous_courseids)
            msg = f"Different course-JSONs with same courseid {course_ids}."
            raise ValidationError(msg)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module test convert."""

from invenio_records_lom.utils import LOMMetadata

from invenio_moodle.convert import convert_moodle_to_lom


def test_convert_moodle_to_lom(
    minimal_record: dict,
    expected_lom_metadata: dict,
) -> None:
    """Test convert_unit_metadata."""
    del minimal_record["courses"][1]

    lom_metadata = LOMMetadata({}, overwritable=True)
    convert_moodle_to_lom(minimal_record, lom_metadata)
    assert lom_metadata.json == expected_lom_metadata



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Convert Moodle metadata format into LOM format."""

from datetime import datetime
from html import unescape

from invenio_records_lom.utils import LOMCourseMetadata, LOMMetadata


class Visitor:
    """Visitor base class."""

    def visit(self, value: dict, record: LOMMetadata) -> None:
        """Convert record from Moodle JSON format to LOMMetadata."""
        for attribute, val in value.items():
            self.visit_attribute(attribute, val, record)

    def visit_attribute(
        self,
        attribute: str,
        value: dict,
        record: LOMMetadata,
    ) -> None:
        """Traverse first level elements of dictionary and extract attributes."""

        def func_not_found(*_: dict, **__: dict) -> None:
            """Ignore keys that are not converted."""

        visit_function = getattr(self, f"visit_{attribute}", func_not_found)
        visit_function(value, record)


class CourseToLOM(Visitor):
    """Course to LOM."""

    def visit_courseid(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit courseid."""
        record.append_identifier(value, catalog="tugrazonline-id")

    def visit_identifier(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit identifier."""
        record.append_identifier(value, catalog="teachcenter-course-id")

    def visit_coursename(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit coursename."""
        self.course_name = value
        record.set_title(value, language_code="x-none")

    def visit_courselanguage(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit courselanguage."""
        record.append_language(value)

    def visit_structure(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit structure."""
        record.append_keyword(value, language_code="x-none")

    def visit_context(self, value: str, record: LOMCourseMetadata) -> None:
        """Visit context."""
        record.append_context(value)

    def visit_objective(self, value: str, record: LOMMetadata) -> None:
        """Visit objective."""
        record.append_description(value, language_code="x-none")

    def visit_description(self, value: str, record: LOMMetadata) -> None:
        """Visit description."""
        record.append_description(unescape(value), language_code="x-none")

    def visit_lecturer(self, value: str, record: LOMMetadata) -> None:
        """Visit lecturer."""
        for lecturer in value.split(","):
            record.append_contribute(lecturer.strip(), role="Author")

    def visit_organisation(self, value: str, record: LOMMetadata) -> None:
        """Visit organisation."""
        record.append_contribute(value, role="Unknown")


class MoodleToLOM(Visitor):
    """Moodle file to lom file."""

    def visit(self, value: dict, record: LOMMetadata) -> None:
        """Visit."""
        self.courses = []

        super().visit(value, record)

        record.set_title(self.title, language_code=self.language)

        version = f"{self.semester} {self.year}"
        record.set_version(version, datetime=self.year)

        for tag in filter(bool, self.tags):
            record.append_keyword(tag, language_code=self.language)

        if abstract := unescape(self.abstract):
            record.append_description(abstract, language_code=self.language)

        for course in self.courses:
            course.set_version(version, datetime=self.year)
            record.append_course(course)

    def visit_courses(self, value: list, _: LOMMetadata) -> None:
        """Visit courses."""
        for course in value:
            lom_course_record = LOMCourseMetadata()
            visitor = CourseToLOM()
            visitor.visit(course, lom_course_record)
            self.courses.append(lom_course_record)

    def visit_year(self, value: str, record: LOMMetadata) -> None:
        """Visit year."""
        self.year = value
        record.set_datetime(value)

    def visit_semester(self, value: str, record: LOMMetadata) -> None:
        """Visit semester."""
        self.semester = value
        record.append_keyword(value, language_code="x-none")

    def visit_language(self, value: str, record: LOMMetadata) -> None:
        """Visit language."""
        self.language = value
        record.append_language(value)

    def visit_title(self, value: str, _: LOMMetadata) -> None:
        """Visit title."""
        self.title = value

    def visit_abstract(self, value: str, _: LOMMetadata) -> None:
        """Visit abstract."""
        self.abstract = value

    def visit_tags(self, value: str, _: LOMMetadata) -> None:
        """Visit tags."""
        self.tags = value

    def visit_persons(self, value: str, record: LOMMetadata) -> None:
        """Visit persons."""
        for person in value:
            name = f"{person['firstname']} {person['lastname']}"
            record.append_contribute(name, role=person["role"])

    def visit_timereleased(self, value: str, record: LOMMetadata) -> None:
        """Visit timereleased."""
        datetime_obj = datetime.fromtimestamp(int(value))
        datetime_isoformat = datetime_obj.date().isoformat()
        record.set_datetime(datetime_isoformat)

    def visit_mimetype(self, value: str, record: LOMMetadata) -> None:
        """Visit mimetype."""
        record.append_format(value)

    def visit_filesize(self, value: str, record: LOMMetadata) -> None:
        """Visit filesize."""
        record.set_size(value)

    def visit_resourcetype(self, value: str, record: LOMMetadata) -> None:
        """Visit resourcetype."""
        # https://skohub.io/dini-ag-kim/hcrt/heads/master/w3id.org/kim/hcrt/slide.en.html
        learningresourcetype_by_resourcetype = {
            "no selection": "other",
            "noselection": "other",
            "presentationslide": "slide",
            "exercise": "assessment",
            "graphic": "image",
            "figure": "image",
        }
        if learningresourcetype := learningresourcetype_by_resourcetype[value.lower()]:
            record.append_learningresourcetype(learningresourcetype)

    def visit_license(self, value: str, record: LOMMetadata) -> None:
        """Visit license."""
        record.set_rights_url(value["source"])

    def visit_classification(self, classifications: str, record: LOMMetadata) -> None:
        """Visit classification."""
        oefos_ids = [
            value["identifier"]
            for classification in classifications
            for value in classification["values"]
        ]

        # reorder to ['1234', '123', '2345', '234', '2']
        oefos_ids.sort(key=lambda id_: id_.ljust(6, chr(255)))

        # metadata.append_oefos(oefos_ids) # noqa: ERA001
        for id_ in oefos_ids:
            record.append_oefos_id(id_)
            record.append_oefos_id(id_, "en")


def convert_moodle_to_lom(moodle_file_metadata: dict, metadata: LOMMetadata) -> None:
    """Convert file metadata."""
    visitor = MoodleToLOM()
    visitor.visit(moodle_file_metadata, metadata)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Provide functions to store files into the database."""

from __future__ import annotations

import hashlib
import re
import typing as t
from pathlib import Path
from tempfile import TemporaryDirectory

from requests import Response, Session

from .types import FileCacheInfo

if t.TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.services import FileService

    from .types import FileRecord


def save_file_locally(
    response: Response,
    hash_md5: str,
    directory: Path,
) -> FileCacheInfo:
    """Save file locally."""
    # find filename in headers
    dispos = response.headers.get("content-disposition", "")

    # all html-headers are encoded in latin1, but python interprets as utf-8
    dispos = dispos.encode("latin1").decode("utf-8")

    if match := re.search('filename="?([^"]*)"?', dispos):
        filename = match.group(1)
    else:
        msg = f"couldn't find filename in header {dispos}"
        raise ValueError(msg)

    # save file to `directory`, compute its hash along the way
    # filepath is of form 'directory/hash_md5/file.pdf'
    filepath = directory.joinpath(hash_md5, filename.replace(" ", "_"))
    hash_ = hashlib.md5()  # noqa: S324

    # create 'directory/0/'
    filepath.parent.mkdir(parents=True)

    with filepath.open(mode="wb", buffering=1024 * 1024) as fp:
        for chunk in response.iter_content(chunk_size=1024 * 1024):
            hash_.update(chunk)
            fp.write(chunk)

    return FileCacheInfo(hash_md5=hash_.hexdigest(), path=filepath)


def cache_file(url: str, hash_md5: str, temp_dir: str) -> FileCacheInfo:
    """Create a file-cache.

    It creates a file-cache by downloading unprovided files into
    `directory` and hashing all files.

    :param Path directory: The directory to download unprovided files into.
    :param list[str] urls: The URLs of the to-be-cached files.
    """
    directory = Path(temp_dir)
    with Session() as session, session.get(url, stream=True) as response:
        response.raise_for_status()

        return save_file_locally(response, hash_md5, directory)


def add_file_to_draft(
    record: FileRecord,
    draft_files: FileService,
    identity: Identity,
) -> None:
    """Add file to draft."""
    with TemporaryDirectory() as temp_dir:
        file_info = cache_file(record.url, record.hash_md5, temp_dir)
        filename = file_info.path.name
        draft_files.init_files(
            id_=record.pid,
            data=[{"key": filename}],
            identity=identity,
        )

        # ATTENTION: 1024 * 1024 may be a problem?
        with file_info.path.open(mode="rb", buffering=1024 * 1024) as fp:
            draft_files.set_file_content(
                id_=record.pid,
                file_key=filename,
                stream=fp,
                identity=identity,
            )
        draft_files.commit_file(id_=record.pid, file_key=filename, identity=identity)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Graz University of Technology.
#
# invenio-moodle is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Api functions."""
from __future__ import annotations

from typing import TYPE_CHECKING

from requests import get
from sqlalchemy.orm.exc import NoResultFound

from .schemas import MoodleSchema
from .utils import (
    build_intermediate_records,
    extract_moodle_records,
    import_record,
    post_processing,
)

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_lom.utils import LOMRecordService


def fetch_moodle(
    moodle_fetch_url: str,
    record_service: LOMRecordService,
    identity: Identity,
) -> None:
    """Fetch data from MOODLE_FETCH_URL and insert it into the database."""
    response = get(moodle_fetch_url, timeout=15)
    response.raise_for_status()

    moodle_data = response.json()

    insert_moodle_into_db(moodle_data, record_service, identity)


def insert_moodle_into_db(
    moodle_data: dict,
    records_service: LOMRecordService,
    identity: Identity,
) -> None:
    """Insert data encoded in `moodle-data` into invenio-database.

    :param dict moodle_data: The data to be inserted into database,
        whose format matches `MoodleSchema`
    :param dict filepaths_by_url: A dictionary
        that maps some file-urls within `moodle_data` to filepaths.
        When a file-url is found in `filepaths_by_url`,
        the file on the corresponding filepath is used.
        Otherwise the file is downloaded from file-url.
    """
    # validate input
    moodle_data = MoodleSchema().load(moodle_data)

    moodle_records = extract_moodle_records(moodle_data)
    post_processing(moodle_records)

    records = build_intermediate_records(moodle_records, records_service, identity)

    for record in records.values():
        try:
            import_record(record, records_service, identity)
        except NoResultFound:  # noqa: PERF203
            continue
